# -*- coding: utf-8 -*-
"""
Package matadata
"""

__version__ = "0.1.6"
__name__ = "infostop"
__author__ = "Ulf Aslak, Laura Alessandretti"
__copyright__ = "Copyright 2020, Ulf Aslak"
__credits__ = ["Ulf Aslak", "Laura Alessandretti"]
__license__ = "MIT"
__maintainer__ = "Ulf Aslak"
__email__ = "ulfjensen@gmail.com"
__status__ = "Development"
